import logging
from typing import Union
from pathlib import Path

import pytest

from ceonstock import filesystem
from ceonstock import errors

logger = logging.getLogger(__name__)

# Folders that store test jobs
THIS_DIR = Path(__file__).parent
JOBS_DIR_SUCCESS = f"{THIS_DIR}/local_jobs/success"
JOBS_DIR_FAIL = f"{THIS_DIR}/local_jobs/fail"

# A single chosen project to test class functions
JOB_PATH = Path(JOBS_DIR_SUCCESS, "Simple Job")


def list_folders_in_dir(directory: Union[str, Path]):
    directory = Path(directory)  # Cast string to Path
    return [entry for entry in directory.iterdir() if entry.is_dir()]


def test_local_job_load():
    local_job = filesystem.CstockLocalJob(JOB_PATH)
    cstock_job = local_job.load()
    logger.info(f"Loaded CstockProject from local_job:\n{cstock_job=}")
    # TODO validate project data?


def test_local_job_load_success():
    source_dir = JOBS_DIR_SUCCESS
    local_job_dirs = list_folders_in_dir(source_dir)
    if not local_job_dirs:
        raise Exception(f"Success cases not tested: No jobs found in {source_dir}")
    for local_job_dir in local_job_dirs:
        logger.info(f"Testing lcal job fail case: {local_job_dir}")
        local_job = filesystem.CstockLocalJob(local_job_dir)
        cstock_job = local_job.load()
        logger.info(f"Loaded CstockProject from local_job:\n{cstock_job=}")


def test_local_job_load_fail_cases():
    source_dir = JOBS_DIR_FAIL
    local_job_dirs = list_folders_in_dir(source_dir)
    if not local_job_dirs:
        raise Exception(f"Failure cases not tested: No jobs found in {source_dir}")
    for local_job_dir in local_job_dirs:
        logger.info(f"Testing lcal job fail case: {local_job_dir}")
        with pytest.raises((errors.CstockInstantiationError, FileNotFoundError, NotADirectoryError)):
            local_job = filesystem.CstockLocalJob(local_job_dir)
            cstock_job = local_job.load()



import logging
import pytest
from typing import Type
from copy import deepcopy

from ceonstock.core.base import CstockBaseClass
from ceonstock.log import printify
from ceonstock import errors

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")


def test_instantiation_success_cases(
    cstock_cls: Type[CstockBaseClass], dicts_to_test: list[dict]
):
    logger.info("Testing success cases...")
    for dict_to_test in dicts_to_test:
        cstock_cls(**dict_to_test)


def test_instantiation_fail_cases(
    cstock_cls: Type[CstockBaseClass], dicts_to_test: list[dict]
):
    logger.info("Testing fail cases...")
    for dict_to_test in dicts_to_test:
        logger.debug(
            f"Instantiating with (fail) dict: {printify(dict_to_test)}"
        )
        # Include TypeError to catch invalid kwargs passed to initiailizer
        with pytest.raises((errors.CstockError, TypeError)):
            cstock_cls(**dict_to_test)


def test_values_for_key(
    cstock_cls: Type[CstockBaseClass],
    base_dict: dict,
    key: str,
    values: list,
):
    """For each item in values_to_test, replace the key_to_test in the base_dict"""
    dict_to_test = deepcopy(base_dict)
    for value in values:
        dict_to_test[key] = value
        cstock_cls(**dict_to_test)



from typing import Dict
from ceonstock import CstockSubmissionInput
from ceonstock.core.submission_input import SubmissionInputEntry


def serialize_entry(submission_input_entry: SubmissionInputEntry) -> Dict:
    as_dict = submission_input_entry.__dict__
    return as_dict


def serialize(cstock_submission_input: CstockSubmissionInput) -> Dict:
    entries_serialized = [
        serialize_entry(entry) for entry in cstock_submission_input.entries
    ]
    as_dict = CstockSubmissionInput.__dict__
    as_dict["entries"] = entries_serialized
    return as_dict



from .local_job import CstockLocalJob
from .local_job import create_local_job
from .local_project import CstockLocalProject
from .path_resolver import PathResolver
from . import filenames
from . import folders
from . import vtt_file



import logging
from ceonstock.core.project_input import CstockProjectInput
from ceonstock.core.file_action import CstockFileAction
from ceonstock.core.job_input import CstockJobInputType
from ceonstock.core.web_input import CstockWebInputType

from . import serialization, instantiation
from .dicts import file_action_dicts

logger = logging.getLogger(__name__)

TEST_CLS = CstockFileAction
TEST_DICTS = file_action_dicts


def test_file_action_dict_serialization() -> None:
    for test_dict in TEST_DICTS.SUCCESS_CASES:
        test_object = TEST_CLS(**test_dict)
        serialization.run_dict_serialization_tests(cstock_object=test_object)


def test_file_action_instantiation_success_cases():
    logger.info("Testing predefined success cases ...")
    instantiation.test_instantiation_success_cases(
        cstock_cls=TEST_CLS, dicts_to_test=TEST_DICTS.SUCCESS_CASES
    )


def test_file_action_instantiation_fail_cases():
    logger.info("Testing predefined fail cases...")
    instantiation.test_instantiation_fail_cases(
        cstock_cls=TEST_CLS, dicts_to_test=TEST_DICTS.FAIL_CASES
    )



import random
import logging
from typing import Optional
from enum import Enum

logger = logging.getLogger(__name__)


def random_selection(options: list, num_values: int = 1, exclude_options=None):
    # Handle exclusion logic here instead of externally so alternative behaviours
    # can be set up when random.sample is out of range (exhausted number of options)
    chosen_values = random.sample(options, num_values)
    return chosen_values


def ordered_selection(
    options: list, num_values: int = 1, start_index=0
) -> list:
    logger.debug(f"Getting ordered selection from ({len(options)}) options.")
    logger.debug(f"{start_index=}")
    chosen_values = options[start_index : num_values + start_index]
    logger.debug(f"{chosen_values=}")
    return chosen_values


class ValuePickerStrategy(Enum):
    ORDERED = "ordered"  # Get the first X found files
    RANDOM = "random"  # Randomly pick from all vailable files


# Handles all internal logic for generating values to be used in project inputs.
# Is cstock job/input agnostic.
# TODO move this and file/data_picker_strategies to another file.
class CstockValuePicker:
    def __init__(
        self,
        available_values: list,
        value_picker_strategy: ValuePickerStrategy = ValuePickerStrategy.RANDOM,
        index_offset: int = 0,
        disable_state: bool = False,
    ):
        """
        A stateful value picker that remembers past choices.
        available_values: The pool of options from which to choose from.
        excluded_values: A list of options that can NOT be picked.
        disable_state: If true, past selections will not be tracked:
            For random selection: Allows the same value to be chosen more than once.
            For ordered selection: Provides a deterministic behaviour in which the same X files will be chosen.
        """
        # self.cstock_project = cstock_project
        self.available_values = available_values
        self.value_picker_strategy = value_picker_strategy
        # TODO on_values_exhausted strategy (reset state, repeat last, error, etc)

        # State config
        self._disable_state = disable_state
        # Caching/memory mechanism
        # A 'memory' for values that have already been chosen, or which are excluded from selection.
        self._previously_picked_values: list = []
        # A 'cursor' to move through an ordered selection.
        self._index_offset: int = index_offset
        logger.debug(f"Value picker instantiated:")
        logger.debug(f"\t{self.available_values=}")
        logger.debug(f"\t{self._disable_state=}")
        logger.debug(f"\t{self._previously_picked_values=}")
        logger.debug(f"\t{self._index_offset=}")

    def _get_valid_values(self):
        logger.debug(f"{self.available_values=}")
        logger.debug(f"{self._previously_picked_values=}")
        valid_values = [
            value
            for value in self.available_values
            if value not in self._previously_picked_values
        ]
        return valid_values

    def get_values(self, num_values=1) -> list:
        """
        Return a list of chosen values.
        """
        logger.debug(f"Getting values {num_values=}")
        valid_values = self._get_valid_values()
        logger.debug(f"{len(valid_values)=}")
        if self.value_picker_strategy == ValuePickerStrategy.RANDOM:
            chosen_values = random_selection(
                valid_values, num_values=num_values
            )
            logger.debug(f"\t{chosen_values=}")
            if not self._disable_state:
                self._previously_picked_values += chosen_values
                logger.debug(f"Updated {self._previously_picked_values=}")
            return chosen_values

        if self.value_picker_strategy == ValuePickerStrategy.ORDERED:
            chosen_values = ordered_selection(
                valid_values,
                num_values=num_values,
                start_index=self._index_offset,
            )
            logger.debug(f"\t{chosen_values=}")
            if not self._disable_state:
                self._index_offset += len(chosen_values)
                logger.debug(f"Updated {self._index_offset=}")
            return chosen_values
        raise Exception(
            f"Invalid value_picker_strategy {self.value_picker_strategy}: Did not return."
        )



# A module to handle converting filename formats between different contexts.
# For example, Houdini uses $F3 to denote 3 digits, while regex uses (TODO what is the regex equiv?)
import logging

logger = logging.getLogger(__name__)


def hou_to_ffmpeg(filename: str):
    replaced = filename.replace("$F", "%0")
    logger.info(f"{filename=}")
    logger.info(f"{replaced=}")

    # Add the 'd' found in ffmpeg glob sysntax
    logger.warning(f"TODO: Use globbing to handle cases where %04d is not 4")
    replaced = replaced.replace("%04", "%04d")

    return replaced


def ffmpeg_to_hou(filename: str):
    replaced = filename.replace("%0", "$F")
    logger.info(f"{filename=}")
    logger.info(f"{replaced=}")
    return replaced



from dataclasses import dataclass
from typing import Dict, List, Union
from pathlib import Path


@dataclass
class SubmissionDirFolderNames:
    entries = ""
    preprocess = "preprocess"
    meta = "meta"
    logs = "logs"


@dataclass
class SubmissionDirs:
    root: Union[Path, str]
    fldr_names: SubmissionDirFolderNames = SubmissionDirFolderNames()

    def __post_init__(self) -> None:
        self.root = Path(self.root)
        self.entries = self.root / self.fldr_names.entries
        self.preprocess = self.root / self.fldr_names.preprocess
        self.meta = self.root / self.fldr_names.meta
        self.logs = self.root / self.fldr_names.logs

    def __str__(self):  # str method determines how this class is shown when printed
        return f"<{self.__class__.__name__}: {self.root}>"

    def dirs(self) -> Dict[str, Path]:
        return {
            "root": self.root,
            "entries": self.entries,
            "preprocess": self.preprocess,
            "meta": self.meta,
            "logs": self.logs,
        }


@dataclass
class JobDirFolderNames:
    job_inputs = "job_inputs"
    job_outputs = "job_outputs"
    meta = "meta"
    logs = "logs"


@dataclass
class JobDirs:
    root: Union[Path, str]
    fldr_names: JobDirFolderNames = JobDirFolderNames()

    def __post_init__(self) -> None:
        self.root = Path(self.root)
        self.job_inputs = self.root / self.fldr_names.job_inputs
        self.job_outputs = self.root / self.fldr_names.job_outputs
        self.meta = self.root / self.fldr_names.meta
        self.logs = self.root / self.fldr_names.logs

    def __str__(self):  # str method determines how this class is shown when printed
        return f"<JobDirs: {self.root}>"

    # TODO deprecate this in favor of to_dict()
    def dirs(self) -> Dict[str, Path]:
        return self.to_dict()

    def to_dict(self) -> Dict[str, Path]:
        return {
            "root": Path(self.root),
            "job_inputs": self.job_inputs,
            "job_outputs": self.job_outputs,
            "logs": self.logs,
            "meta": self.meta,
        }


def create_dirs(dirs_to_create: List[Path]) -> List[Path]:
    new_dirs = []
    for path in dirs_to_create:
        if not path.exists():
            Path(path).mkdir(parents=True)
            new_dirs.append(path)
    return new_dirs



from __future__ import annotations
import logging
from dataclasses import dataclass, field
from typing import ClassVar, Type, Any, List, Dict
from typing import Optional, Union
from typing import TypedDict

from . import CstockProjectInput
from .base import CstockBaseClass
from .file_action import CstockFileAction

logger = logging.getLogger(__name__)


# class CropSettings(TypedDict):
#     x: float
#     y: float
#     width: float
#     height: float


@dataclass
class CstockSubmissionInputRaw:
    """
    Raw inputs represent the files/values as directly received from the web-client.
    Some files may need to be pre-processed e.g. convert a word document to a set of images.
    """

    values: List
    project_input: CstockProjectInput
    preprocess_actions: List[CstockFileAction] = field(default_factory=list)

    def __post_init__(self):
        # Enforce loading of types when loaded from a dict (imported json)
        if not isinstance(self.project_input, CstockProjectInput):
            self.project_input = CstockProjectInput(**self.project_input)

        def enforce_file_action(file_action) -> CstockFileAction:
            if isinstance(file_action, CstockFileAction):
                return file_action
            # If it's not a CstockFileAction, assume it's a dict and try to load
            return CstockFileAction(**file_action)

        self.preprocess_actions = [
            enforce_file_action(file_action)
            for file_action in self.preprocess_actions
        ]

    def __str__(self):
        return f"<{self.__class__.__name__} for '{self.project_input.name}': {self.values}>"


@dataclass
class CstockSubmissionInputEntry(CstockBaseClass):
    value: Any
    entry_actions: List[CstockFileAction]

    def __post_init__(self):
        # Enforce class isntances when loaded from a dict
        def enforce_instance(
            entry_action: Union[dict, CstockFileAction]
        ) -> CstockFileAction:
            if isinstance(entry_action, CstockFileAction):
                return entry_action
            return CstockFileAction(**entry_action)

        self.entry_actions = [
            enforce_instance(action) for action in self.entry_actions
        ]


@dataclass
class CstockSubmissionInput(CstockBaseClass):
    """Represents a single submission for a project input.
    A single submission can contain multiple entries.
    Each entry may contain extra data depending on its type.
    For example, images can optionally contain information about how to crop them.
    """

    entries: List[CstockSubmissionInputEntry]
    project_input: CstockProjectInput

    def __post_init__(self):
        # Handle loading from json/serializable dict.
        if isinstance(self.project_input, dict):
            self.project_input = CstockProjectInput(**self.project_input)

        def enforce_instance(
            entry: Union[Dict, CstockSubmissionInputEntry]
        ) -> CstockSubmissionInputEntry:
            if isinstance(entry, CstockSubmissionInputEntry):
                return entry
            return CstockSubmissionInputEntry(**entry)

        self.entries = [enforce_instance(entry) for entry in self.entries]

    def __str__(self):
        msg = f"<{self.__class__.__name__} '{self.project_input.name}': {len(self.entries)} entries>"
        return msg



from ceonstock import CstockJobInputType, CstockWebInputType

# Examples that should always work
SUCCESS_CASES = [
    {
        "name": "myTestJobInput",
        "job_input_type": CstockJobInputType.IMG,
        "values": ["folder/to/somefile.jpg"],
    },
    {  # Empty values list
        "name": "myTestJobInput",
        "job_input_type": CstockJobInputType.IMG,
        "values": [],
    },
    {  # string instead of enum (lowercase)
        "name": "myTestJobInput",
        "job_input_type": "img",
        "values": ["rel_fldr/to/somefile.jpg"],
    },
    {  # string instead of enum (uppercase)
        "name": "myTestJobInput",
        "job_input_type": "IMG",
        "values": ["rel_fldr/to/somefile.jpg"],
    },
]

# Examples that we know should not work.
FAIL_CASES = [
    {  # Fail unrecognized kwargs to alert user of possible unexpected data inputs if unpacking a dict
        "name": "myTestJobInput",
        "job_input_type": CstockJobInputType.IMG,
        "values": [],
        "unknown_kwarg": [1, 2, 3],
    },
    {  # Fail if 'values' is not a list
        "name": "myTestJobInput",
        "job_input_type": CstockJobInputType.IMG,
        "values": "res_fldr_to_somefile.jpg",
    },
    {  # non-string value instead of enum
        "name": "myTestJobInput",
        "job_input_type": 1,
        "values": ["rel_fldr/to/somefile.jpg"],
    },
    {  # string instead of enum (unrecognized string)
        "name": "myTestJobInput",
        "job_input_type": "TypeThatDoesntExist",
        "values": ["rel_fldr/to/somefile.jpg"],
    },
    {  # Passed the wrong Enum type to job_input (the passed enum value 'dropdown' doesn't exist on the expected enum)
        "name": "myTestJobInput",
        "job_input_type": CstockWebInputType.DROPDOWN,
        "values": ["rel_fldr/to/somefile.jpg"],
    },
    {  # Passed the wrong Enum type to job_input (but the passed enum value 'img' DOES also exist on the expected enum!)
        "name": "myTestJobInput",
        "job_input_type": CstockWebInputType.IMG,
        "values": ["rel_fldr/to/somefile.jpg"],
    },
]



import logging
from ceonstock import CstockJob, CstockJobType

from . import serialization, instantiation
from .dicts import job_dicts

logger = logging.getLogger(__name__)

TEST_CLS = CstockJob
TEST_DICTS = job_dicts


def test_job_dict_serialization() -> None:
    for test_dict in TEST_DICTS.SUCCESS_CASES:
        test_object = TEST_CLS(**test_dict)
        serialization.run_dict_serialization_tests(cstock_object=test_object)


def test_job_instantiation_success_cases():
    logger.info("Testing predefined success cases ...")
    instantiation.test_instantiation_success_cases(
        cstock_cls=TEST_CLS, dicts_to_test=TEST_DICTS.SUCCESS_CASES
    )

    logger.info("Testing init with all known CstockJobTypes ...")
    instantiation.test_values_for_key(
        cstock_cls=TEST_CLS,
        base_dict=TEST_DICTS.SUCCESS_CASES[0],
        key="job_type",
        values=list(CstockJobType),
    )

    logger.warning("TODO validate that values are appropriate for job_type")


def test_job_instantiation_fail_cases():
    logger.info("Testing predefined fail cases...")
    instantiation.test_instantiation_fail_cases(
        cstock_cls=TEST_CLS, dicts_to_test=TEST_DICTS.FAIL_CASES
    )



from ceonstock.core.base import CstockBaseEnum


class CstockWebInputType(CstockBaseEnum):
    AUDIO = "audio"
    BOOL = "bool"
    COLOR = "color"
    DOC = "doc"  # Convert docs to imgs
    DROPDOWN = "dropdown"
    IMG = "img"
    INT = "int"
    TEXT = "text"


# TODO move metadata / info modules elsewhere
# @dataclass(frozen=True)
# class WebInputTypeInfo:
#     """Information about a particular CstockWebInputType which does NOT change per-instance"""

#     job_input_type: job_input.CstockJobInputType
#     is_file: bool = False


# def get_info(web_input_type: CstockWebInputType):
#     LOOKUP = {
#         CstockWebInputType.AUDIO: WebInputTypeInfo(
#             job_input_type=job_input.CstockJobInputType.AUDIO,
#             is_file=True,
#         ),
#         CstockWebInputType.COLOR: WebInputTypeInfo(
#             job_input_type=job_input.CstockJobInputType.COLOR,
#         ),
#         CstockWebInputType.DOC: WebInputTypeInfo(
#             job_input_type=job_input.CstockJobInputType.IMG,
#             is_file=True,
#         ),
#         CstockWebInputType.IMG: WebInputTypeInfo(
#             job_input_type=job_input.CstockJobInputType.IMG,
#             is_file=True,
#         ),
#     }
#     try:
#         type_class = LOOKUP[web_input_type]
#     except KeyError:
#         raise errors.CstockUnknownTypeError(
#             received=web_input_type, expected=CstockWebInputType
#         )
#     return type_class



# A custom Formatter that injects extra variables that can be used in the format string.
# In this case we add task_name and task_id.
import logging
import time

from enum import Enum

from .colors import TerminalColor


# Modify the default formatter to inject extra parameters
# which can be used in the format string
class CeleryTaskFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            from celery._state import get_current_task

            self.get_current_task = get_current_task
        except ImportError:
            self.get_current_task = lambda: None

    def formatTime(self, record, datefmt=None):
        """Processes the %(asctime)s arg in the format string.
        Modified to include the use of %f and %z for milliseconds/timezone
        in the datefmt string"""
        ct = self.converter(record.created)
        if datefmt:
            # support %z and %f in datefmt (struct_time doesn't carry ms or tz)
            datefmt = datefmt.replace("%f", "%03d" % int(record.msecs))
            datefmt = datefmt.replace("%z", time.strftime("%z"))
            s = time.strftime(datefmt, ct)
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s,%03d" % (t, record.msecs)
        return s

    def format(self, record):
        # Modify the recored.__dict__ to add extra properties that can
        # be used for formatting.
        task = self.get_current_task()
        if task and task.request:
            task_kwargs = task.request.kwargs
            job_uuid = task_kwargs.get("job_uuid")
            task_name = task.name
            # Trim the taskname by lstripping redundant parts: app.task.
            task_name = _lstrip_substring(task_name, "app.")
            task_name = _lstrip_substring(task_name, "tasks.")
            task_name = task_name.lstrip(".")
            record.__dict__.update(
                task_id=task.request.id, task_name=task_name, job_uuid=job_uuid
            )
        else:
            record.__dict__.setdefault("task_name", "")
            record.__dict__.setdefault("task_id", "")
            record.__dict__.setdefault("job_uuid", "")
        return super().format(record)


def _lstrip_substring(string_to_check: str, substring_to_remove: str):
    if not string_to_check.startswith(substring_to_remove):
        return string_to_check
    return string_to_check[len(substring_to_remove) :]


# Wrap the task formatter with colouring functionality.
class CeleryTaskFormatterColored(CeleryTaskFormatter):
    default_format = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # def __init__(self, string_for_format: str, *args, **kwargs):
    #     self.msg_format = TaskFormatterColored.default_format
    #     if string_for_format:
    #         self.msg_format = string_for_format
    #     # super().__init__(self.msg_format, *args, **kwargs)

    def get_color(self, levelno):
        COLORS = {
            logging.DEBUG: TerminalColor.BLUE,
            logging.INFO: TerminalColor.GREY,
            logging.WARNING: TerminalColor.YELLOW,
            logging.ERROR: TerminalColor.RED,
            logging.CRITICAL: TerminalColor.RED_BOLD,
        }
        chosen_color = COLORS.get(levelno, "")
        return chosen_color

    def apply_color(self, formatted_msg: str, terminal_color: TerminalColor):
        # The color formatting must be applied to each line individually to take effect
        # over newlines.
        colored_lines = []
        for line in formatted_msg.split("\n"):
            colored_line = f"{terminal_color.value}{line}{TerminalColor._RESET.value}"
            colored_lines.append(colored_line)
        colored_msg = "\n".join(colored_lines)
        return colored_msg

    def format(self, record):
        formatted_msg = super().format(record)
        chosen_color = self.get_color(record.levelno)
        colored_msg = self.apply_color(formatted_msg, chosen_color)
        return colored_msg


# class TaskFormatterColored(formatters.ColoredFormatter, TaskFormatter):
#     pass


""" Example Usage:
logger = logging.getLogger()
sh = logging.StreamHandler()
sh.setFormatter(
    TaskFormatter(
        '%(asctime)s - %(task_id)s - %(task_name)s - %(name)s - %(levelname)s - %(message)s'))
logger.setLevel(logging.INFO)
logger.addHandler(sh)
"""



import logging
import pytest
from typing import Union
from uuid import UUID, uuid4
from pathlib import Path
from ceonstock import CstockProjectInput
from ceonstock import json_io
from ceonstock import errors

logger = logging.getLogger(__name__)


def _get_json_files_in_dir(dir_to_read: Union[str, Path]) -> list[Path]:
    if not Path(dir_to_read).is_dir():
        raise Exception(
            f"Could not get test files, dir does not exist: {dir_to_read}"
        )
    p = Path(dir_to_read).glob("**/*")
    files = [x for x in p if x.is_file()]
    if not files:
        raise Exception(f"No files found for testing in path: {dir_to_read}")
    return files


TEST_CLS = CstockProjectInput
JSON_FILES_DIR = f"{Path(__file__).parent}/json_files/project_input"
FILE_DIR_SUCCESS = f"{JSON_FILES_DIR}/success"
FILE_DIR_FAILURE = f"{JSON_FILES_DIR}/failure"
FILES_TO_TEST_SUCCESS = _get_json_files_in_dir(FILE_DIR_SUCCESS)
FILES_TO_TEST_FAILURE = _get_json_files_in_dir(FILE_DIR_FAILURE)


def test_render_task_load_from_json_success() -> None:
    logger.info(f"Loading files: {FILES_TO_TEST_SUCCESS}")
    for test_file in FILES_TO_TEST_SUCCESS:
        logger.info(f"Loading file: {test_file}")
        json_data = json_io.load(test_file)
        test_object = TEST_CLS(**json_data)


def test_render_task_load_from_json_failure() -> None:
    logger.info(f"Loading files: {FILES_TO_TEST_FAILURE}")
    for test_file in FILES_TO_TEST_FAILURE:
        with pytest.raises((TypeError, errors.CstockUnknownTypeError)):
            logger.info(f"Loading file: {test_file}")
            json_data = json_io.load(test_file)
            test_object = TEST_CLS(**json_data)



import logging
from typing import Type, Optional
from pathlib import Path

from ceonstock.core.base import CstockBaseClass
from ceonstock.log import printify
from ceonstock import json_io

from . import helpers

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

THIS_DIR = Path(__file__).parent
FILE_EXPORT_DIR = f"{THIS_DIR}/json_files/serialization_exports"


def json_to_cstock(
    json_string, cstock_cls: Type[CstockBaseClass]
) -> CstockBaseClass:
    """Loads a json string into a cstock object"""
    logger.debug(f"Converting json to {cstock_cls}...")
    dict_data = json_io.loads(json_string)
    logger.debug(f"Loaded dict_data: {dict_data}")
    # Instantiate cstock_object from dict
    cstock_object = cstock_cls(**dict_data)
    logger.debug(f"Created cstock_object: {cstock_object}")
    return cstock_object


def cstock_to_json(
    cstock_object: CstockBaseClass, write_file: Optional[Path] = None
) -> str:
    logger.debug(f"Converting {cstock_object} to json...")
    dict_data = cstock_object.__dict__
    logger.debug(f"Received dict_data:\n{dict_data}")
    json_string = json_io.dumps(dict_data)
    logger.debug(f"Created json string:\n{json_string}")
    if write_file:
        logger.info(f"Saving JSON to file: {write_file}")
        with open(write_file, "w") as f:
            f.write(json_string)
    return json_string


def test_json_to_cstock(json_string: str, cstock_cls: Type[CstockBaseClass]):
    """Load a cstock object from a json string.
    Convert the new cstock object back to a JSON string and check for equality
    """

    # Load a cstock object from a JSON string
    logger.debug("Converting json to cstock ...")
    cstock_object = json_to_cstock(json_string, cstock_cls)
    # Then convert it back to JSON
    logger.debug("Converting new cstock object back to json ...")
    back_to_json = cstock_to_json(cstock_object)

    # Check for equality
    logger.debug(
        f"\noriginal / cstock / back_to_json\
    \n\t{json_string}\
    \n\t{cstock_object}\
    \n\t{back_to_json}"
    )
    assert json_string == back_to_json


def test_cstock_to_json(cstock_object: CstockBaseClass):
    """Convert a Cstock object to a JSON string.
    Load a cstock object from the JSON string and check for equality"""
    # Load a cstock object from a JSON string
    logger.debug("Converting cstock to JSON ...")
    json_string = cstock_to_json(cstock_object)
    # Then convert it back to JSON
    logger.debug(
        f"Converting json string back to {cstock_object.__class__} ..."
    )
    back_to_cstock = json_to_cstock(json_string, cstock_object.__class__)

    # Check for equality
    logger.debug(
        f"\noriginal / json / back_to_cstock\
    \n\t{cstock_object}\
    \n\t{json_string}\
    \n\t{back_to_cstock}"
    )
    assert cstock_object.__dict__ == back_to_cstock.__dict__


def run_json_serialization_tests(cstock_object: CstockBaseClass):
    # Export the json string to a file for inspection
    filename = f"{cstock_object.__class__.__name__}.json"
    filepath = Path(FILE_EXPORT_DIR, filename)
    json_string = cstock_to_json(cstock_object, write_file=filepath)

    # Begin tests
    test_cstock_to_json(cstock_object)
    test_json_to_cstock(json_string, cstock_object.__class__)



from . import job_input_dicts
from uuid import uuid4

SUCCESS_CASES = [
    {  # Without job_uuid
        "job_inputs": [job_input_dicts.SUCCESS_CASES[0]],
        "job_type": "preview",
        # Valid but non-existant uuid
        "project_uuid": str(uuid4()),
        # No job_uuid provided
    },
    {  # With job_uuid
        "job_inputs": [job_input_dicts.SUCCESS_CASES[0]],
        "job_type": "preview",
        # Valid but non-existant uuid
        "project_uuid": str(uuid4()),
        "job_uuid": str(uuid4())
        # No job_uuid provided
    },
]

FAIL_CASES = [
    {"job_type": "preview", "project_uuid": str(uuid4())}  # Missing job inputs
]



import logging
from typing import Union
from pathlib import Path

import pytest

from ceonstock import filesystem

logger = logging.getLogger(__name__)

# Folders that store test jobs
THIS_DIR = Path(__file__).parent


class FilenameTest:
    def __init__(self, in_filename: str, expected_output: str):
        self.in_filename = in_filename
        self.expected_output = expected_output


filename_tests_hou_to_ffmpeg = [
    FilenameTest("myrender.$F4.exr", "myrender.%04d.exr"),
]

filename_tests_ffmpeg_to_hou = [
    FilenameTest("myrender.%04.exr", "myrender.$F4.exr"),
]


def test_hou_to_ffmpeg():
    for testcase in filename_tests_hou_to_ffmpeg:
        result = filesystem.filenames.hou_to_ffmpeg(testcase.in_filename)
        assert result == testcase.expected_output


def test_ffmpeg_to_hou():
    for testcase in filename_tests_ffmpeg_to_hou:
        result = filesystem.filenames.ffmpeg_to_hou(testcase.in_filename)
        assert result == testcase.expected_output



# Helper functions for working with color
from typing import Tuple


# TODO parse hex string to produce value
def color_hex_to_rgb(hex_value: str) -> Tuple[int, int, int]:
    """Convert from a hexstring to a list of floats in the 0-1 range"""
    hex_value = hex_value.lstrip("#")
    # print(f"Got hex color: {hex_value}")
    if len(hex_value) == 3:
        # If we receive a 3-char hex string, double it to accurately represent the full hex col
        # E.g. #F6B Should become #FF66BB
        hex_value = "".join([char * 2 for char in hex_value])
    # print(f"Processing hex color: {hex_value}")
    # Slice the string into groups of 2 chars and convert to int.
    rgb = tuple(int(hex_value[i : i + 2], 16) for i in (0, 2, 4))
    # TODO rgba version (in a separate function)
    # rgba = tuple(int(hex_value[i : i + 2], 16) for i in (0, 2, 4, 6))
    return rgb


def color_hex_to_hou(hex_value: str) -> Tuple[float, float, float]:
    rgb_color = color_hex_to_rgb(hex_value)
    normalized_color = tuple([rgb_to_normalized(val) for val in rgb_color])
    return normalized_color


def rgb_to_normalized(rgb_value: int) -> float:
    """Convert from 0-255(int) to 0-1(float)"""
    normalized_value = rgb_value / 255.0
    return normalized_value


def normalized_to_rgb(normalized_value: float) -> int:
    """Convert from 0-1 (float) to 0-255 (int)"""
    rgb_value = round(normalized_value * 255)
    return rgb_value



from uuid import uuid4
from ceonstock.core import CstockRenderAppType
from ceonstock.core import render_task

SUCCESS_CASES = [
    {  # Expected case
        "task_name": "myTestRenderTask",
        "app_type": CstockRenderAppType.HOU,
        "app_version": "19.5",
        "app_render_settings": render_task.CstockRenderTaskAppSettingsHou(
            frames="0 10", target_node="/out/karma1"
        ),
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_dependencies": [],
    },
    {  # App from string
        "task_name": "myTestRenderTask",
        "app_type": "hou",
        "app_version": "19.5",
        "app_render_settings": render_task.CstockRenderTaskAppSettingsHou(
            frames="0 10", target_node="/out/karma1"
        ),
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_dependencies": [],
    },
    {  # Render settings from dict
        "task_name": "myTestRenderTask",
        "app_type": CstockRenderAppType.HOU,
        "app_version": "19.5",
        "app_render_settings": {
            "frames": "0 10",
            "target_node": "/out/karma1",
        },
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_dependencies": [],
    },
    {  # With UUID object
        "task_name": "myTestRenderTask",
        "app_type": CstockRenderAppType.HOU,
        "app_version": "19.5",
        "app_render_settings": render_task.CstockRenderTaskAppSettingsHou(
            frames="0 10", target_node="/out/karma1"
        ),
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_uuid": uuid4(),
        "task_dependencies": [],
    },
    {  # With UUID (as string)
        "task_name": "myTestRenderTask",
        "app_type": CstockRenderAppType.HOU,
        "app_version": "19.5",
        "app_render_settings": render_task.CstockRenderTaskAppSettingsHou(
            frames="0 10", target_node="/out/karma1"
        ),
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_uuid": str(uuid4()),
        "task_dependencies": [],
    },
]

FAIL_CASES = [
    {  # Invalid app string
        "task_name": "myTestRenderTask",
        "app_type": "fakeapp_type",
        "app_version": "19.5",
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "path/to/otuputfile/todo/figure_out_proper_approach",
        "task_dependencies": [],
    },
    {  # With invalid uuid string (non-hex chars)
        "task_name": "myTestRenderTask",
        "app_type": CstockRenderAppType.HOU,
        "app_version": "19.5",
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_uuid": "invalidwithillegalchars",
        "task_dependencies": [],
    },
    {  # With invalid uuid string (too long) (but using only valid hex chars)
        "task_name": "myTestRenderTask",
        "app_type": CstockRenderAppType.HOU,
        "app_version": "19.5",
        "task_input": "relpath/to/somefile/my_project_hip.hiplc",
        "task_output": "relpath/to/otuputfile/todo/figure_out_proper_approach",
        "task_uuid": "0a47c1dc-47e1-4c22-ac51-0800bd5024c63",
        "task_dependencies": [],
    },
]



from .exporter import dump, dumps, save_json
from .importer import load, loads
from .importer import cstock_from_json
from . import errors

# from .file_handler import write_json_file, read_json_file



import logging

from ceonstock import CstockProjectInput
from ceonstock import CstockJobInput
from ceonstock import CstockJobInputType
from ceonstock.core.web_input import CstockWebInputType

from ceonstock.job_maker import value_picker
from ceonstock.job_maker import job_input_maker

logger = logging.getLogger(__name__)


JOB_INPUT_FILES_DIR = "/mnt/FileStorage/Dayne/Web/proj_ceonstock/assets/project_input_files/job_test_rendering_files"

PROJECT_INPUTS_TO_TEST = [
    CstockProjectInput(
        name="single_image",
        job_input_type=CstockJobInputType.IMG,
        web_input_type=CstockProjectInput.web_input_types.IMG,
        description="A single image TODO system to enforce limit on num inputs",
    ),
    CstockProjectInput(
        name="image_from_doc",
        job_input_type=CstockJobInputType.IMG,
        web_input_type=CstockProjectInput.web_input_types.DOC,
        description="A single image TODO system to enforce limit on num inputs",
    ),
    CstockProjectInput(
        name="single_color",
        job_input_type=CstockJobInputType.COLOR,
        web_input_type=CstockProjectInput.web_input_types.COLOR,
        description="A single color",
    ),
    CstockProjectInput(
        name="dropdown_string",
        job_input_type=CstockJobInputType.STRING,
        web_input_type=CstockProjectInput.web_input_types.DROPDOWN,
        type_settings={
            "dropdown_options": [
                {
                    "value": "dropdown_option_value_0",
                    "label": "Option Label 0",
                },
                {
                    "value": "dropdown_option_value_1",
                    "label": "Option Label 1",
                },
                {
                    "value": "dropdown_option_value_2",
                    "label": "Option Label 2",
                },
            ]
        },
        description="A single image TODO system to enforce limit on num inputs",
    ),
]


def test_create_job_inputs_random_inputs():
    vp = value_picker.JobInputValuePickerFromFilesRandom(JOB_INPUT_FILES_DIR)
    for project_input in PROJECT_INPUTS_TO_TEST:
        new_job_input = job_input_maker.create_job_input(project_input, vp)
        logger.info(f"New job input: {new_job_input}")
    logger.warning("TODO assert expected job_input state")


def test_create_job_inputs_ordered_inputs():
    vp = value_picker.JobInputValuePickerFromFilesOrdered(JOB_INPUT_FILES_DIR)
    for project_input in PROJECT_INPUTS_TO_TEST:
        new_job_input = job_input_maker.create_job_input(project_input, vp)
        _validate_job_input(project_input, new_job_input)
        # TODO global mechanism for validation
        logger.info(f"New job input: {new_job_input}")
    logger.warning("TODO assert expected job_input state")


# TODO global mechanism for validation
def _validate_job_input(
    cstock_project_input: CstockProjectInput, cstock_job_input: CstockJobInput
):
    logger.warning("TODO: Global mechanism for ceonstock data validation")
    if cstock_project_input.web_input_type == CstockWebInputType.DROPDOWN:
        assert len(cstock_job_input.values) == 1



import pytest

# we want to have pytest assert introspection in the helpers
# Otherwise, only asserts directly inside the currently
# running .py test file will print debug info
pytest.register_assert_rewrite("ceonstock.json_io.tests.serialization")
pytest.register_assert_rewrite("ceonstock.json_io.tests.json_tests")



import logging
from uuid import UUID, uuid4
from ceonstock.core import (
    CstockRenderTask,
    CstockRenderAppType,
    CstockRenderPipeline,
)
from . import serialization, instantiation
from .dicts import render_pipeline_dicts

logger = logging.getLogger(__name__)

TEST_CLS = CstockRenderPipeline
TEST_DICTS = render_pipeline_dicts


def test_render_task_dict_serialization() -> None:
    for test_dict in TEST_DICTS.SUCCESS_CASES:
        test_object = TEST_CLS(**test_dict)
        serialization.run_dict_serialization_tests(cstock_object=test_object)


def test_render_task_instantiation_success_cases() -> None:
    logger.info("Testing predefined success cases ...")
    instantiation.test_instantiation_success_cases(
        cstock_cls=TEST_CLS, dicts_to_test=TEST_DICTS.SUCCESS_CASES
    )


# TODO setup proper fail cases (currently there is no validation)
# def test_render_task_instantiation_fail_cases() -> None:
#     logger.info("Testing predefined fail cases...")
#     instantiation.test_instantiation_fail_cases(
#         cstock_cls=TEST_CLS, dicts_to_test=TEST_DICTS.FAIL_CASES
#     )



# Import/Export for a ceonstock Project
# To avoid polluting the base classes implement (de)serialization tools in this module directly,
import json
import logging
from uuid import uuid4
from dataclasses import asdict

from ceonstock.core import CstockProject, web_input
from ceonstock.core.project import CstockProjectInfo
from ceonstock.core.project_input import CstockProjectInput
from ceonstock import CstockJobInputType
from ceonstock import CstockWebInputType
from ceonstock import constants

from ceon_render import (
    CeonFileReference,
    CeonFileSourceType,
    CeonRenderPipeline,
    CeonRenderPipelineJob,
)

logger = logging.getLogger(constants.LOGGER_NAME)

# POSSIBLE ALTERNATIVE dataclass-wizard.
#    https://dataclass-wizard.readthedocs.io/en/latest/
# dataclass_wizard:
# - Expects enums to use VALUES by default: 'color' not 'COLOR'
# - Only provider one error at a time on failure.
# - Doesn't require schema class, just use fromdict and asdict methods directly.
# - Good documentation and seems to be easy to config

# from dataclass_wizard import fromdict, asdict, DumpMeta


# marshmallow:
# - Gives cleaerer errors (reports all together)
# - Expects enums to use NAME not VALUE, e.g. 'COLOR' not 'color'
# - Requires instantiating the schema class first

# Chose to use marshmallow for the multi-error reporting. NOTE: enum NAME vs VALUE could cause
# some compatibility problems with workflows that expect values? It's possible to configures
# marshmallow but haven't looked into it yet and seems to not support enums contained
# within a list?

# Keep dataclass_wizard available as a backup in case the enum naming convention causes problems in the
# future? (Or learn how to configure marshmallow to use enum values)

# import marshmallow_dataclass

# Create a marshmallow schema instance to use for (de)serialization operations.
# project_schema = marshmallow_dataclass.class_schema(CstockProject)()


def _project_input_from_dict(dict_data) -> CstockProjectInput:
    prepared_dict = dict_data
    try:
        # Load from VALUE
        job_input_type = CstockJobInputType(dict_data["job_input_type"])
    except ValueError:
        # Load from NAME
        job_input_type = CstockJobInputType[dict_data["job_input_type"]]
    prepared_dict["job_input_type"] = job_input_type
    prepared_dict["web_input_type"] = CstockWebInputType(dict_data["web_input_type"])
    return CstockProjectInput(**prepared_dict)


def _render_pipeline_job_from_dict(dict_data) -> CeonRenderPipelineJob:
    prepared_dict = dict_data.copy()
    job_input = dict_data["job_input"]

    # Only pass reference type if speicified in dict to prevent overwriting class
    # defaults
    file_reference_args = {"target": job_input["target"]}
    file_reference_type = job_input.get("file_source")
    if file_reference_type:
        file_reference_args["file_source"] = CeonFileSourceType(file_reference_type)
    file_reference = CeonFileReference(**file_reference_args)
    prepared_dict["job_input"] = file_reference

    return CeonRenderPipelineJob(**prepared_dict)


def _render_pipeline_from_dict(dict_data: dict) -> CeonRenderPipeline:
    prepared_dict = dict_data.copy()
    prepared_dict["pipeline_jobs"] = [
        _render_pipeline_job_from_dict(data) for data in dict_data["pipeline_jobs"]
    ]
    return CeonRenderPipeline(**prepared_dict)


def from_dict(dict_data: dict) -> CstockProject:
    """Convert a JSON dict (simple types only) into a CstockProject instance"""
    prepared_dict = dict_data.copy()
    prepared_dict["project_inputs"] = [
        _project_input_from_dict(data) for data in dict_data["project_inputs"]
    ]
    prepared_dict["render_pipeline"] = _render_pipeline_from_dict(
        dict_data["render_pipeline"]
    )
    prepared_dict["project_info"] = CstockProjectInfo(**dict_data["project_info"])
    return CstockProject(**prepared_dict)


def from_file(file_path: str) -> CstockProject:
    with open(file_path, "r") as f:
        data = json.load(f)
    project_id = data.get("project_id")
    if not project_id:
        new_id = uuid4()
        logger.warning(
            "Loaded json file does not contain a project id (%s). Writing to json file new id: %s",
            file_path,
            new_id,
        )
        data["project_id"] = str(uuid4())
        with open(file_path, "w") as f:
            f.write(json.dumps(data, indent=4, sort_keys=True))
        logger.warning("Rewrote json file: %s", file_path)
    project = from_dict(data)
    return project


def to_file(file_path: str, project: CstockProject) -> str:
    """Save a Cstockproject to a json file.
    Returns: The path of the newly saved file as a string"""
    project_dict = asdict(project)
    with open(file_path, "w") as f:
        json.dump(project_dict, f, indent=2)
    return file_path



import logging
from ceonstock import CstockJobInput, CstockJobInputType

from . import json_tests
from ceonstock.tests.dicts import job_input_dicts

logger = logging.getLogger(__name__)

TEST_CLS = CstockJobInput
TEST_DICTS = job_input_dicts


def test_job_input_json_serialization() -> None:
    for test_dict in TEST_DICTS.SUCCESS_CASES:
        test_object = TEST_CLS(**test_dict)
        json_tests.run_json_serialization_tests(cstock_object=test_object)



from typing import Union, Type, Optional, List
from pathlib import Path

# from ceonstock.errors import CstockError


class EmptyJSONFileError(Exception):
    """Raised when a JSON file is found, but does not contain any content"""


class InvalidJSONFileError(Exception):
    """Raised when a JSON file is found, but could not be decoded"""

    def __init__(
        self, file_path: Union[str, Path], errors: Optional[list] = None
    ):
        self.file_path = str(file_path)
        self.errors = errors
        self.message = f"Failed to read json file {file_path}"
        if errors:
            self.message += ": " + str(errors)
        super().__init__(self.message)


# class CstockImportError(CstockError):
#     """Raised when trying to load an object from json data"""

#     def __init__(
#         self,
#         file_path: Union[str, Path],
#         class_to_instantiate: Type,
#         errors: Optional[List] = None,
#     ):
#         self.file_path = str(file_path)
#         self.class_to_instantiate = class_to_instantiate.__name__
#         self.errors = errors
#         self.message = (
#             f"Failed to import class {class_to_instantiate.__name__} from {file_path}"
#         )
#         if errors:
#             self.message += ": " + str(errors)
#         super().__init__(self.message)
